@@ -1,5 +1,5 @@
 #####
-# Copyright (c) 2011-2012, NVIDIA Corporation.  All rights reserved.
+# Copyright (c) 2011-2014, NVIDIA Corporation.  All rights reserved.
 #
 # Redistribution and use in source and binary forms, with or without
 # modification, are permitted provided that the following conditions are met:
@@ -30,17 +30,13 @@
 # nvidia_smi
 # nvml_bindings <at> nvidia <dot> com
 #
-# Sample code that attempts to reproduce the output of nvidia-smi -q- x
+# Sample code that attempts to reproduce the output of nvidia-smi -q -x
 # For many cases the output should match
 #
+# Can be used as a library or a command line script
+#
 # To Run:
-# $ python
-# Python 2.7 (r27:82500, Sep 16 2010, 18:02:00) 
-# [GCC 4.5.1 20100907 (Red Hat 4.5.1-3)] on linux2
-# Type "help", "copyright", "credits" or "license" for more information.
-# >>> import nvidia_smi
-# >>> print(nvidia_smi.XmlDeviceQuery())
-# ...
+# $ python nvidia_smi.py
 #
 
 from pynvml import *
@@ -116,6 +112,44 @@ def GetEccStr(handle):
     strResult += '      </aggregate>\n'
     return strResult
 
+def GetRetiredPagesByCause(handle, cause):
+    strResult = ''
+    try:
+        pages = nvmlDeviceGetRetiredPages(handle, cause)   
+        count = str(len(pages))
+    except NVMLError as err:
+        error = handleError(err)
+        pages = None
+        count = error
+    strResult += '        <retired_count>' + count + '</retired_count>\n'
+    if pages is not None:
+        strResult += '        <retired_page_addresses>\n'
+        for page in pages:
+            strResult += '          <retired_page_address>' + "0x%016x" % page + '</retired_page_address>\n'
+        strResult += '        </retired_page_addresses>\n'
+    else:
+        strResult += '        <retired_page_addresses>' + error + '</retired_page_addresses>\n'
+    return strResult
+
+def GetRetiredPagesStr(handle):
+    strResult = ''
+    causes = [ "multiple_single_bit_retirement", "double_bit_retirement" ]
+    for idx in range(NVML_PAGE_RETIREMENT_CAUSE_COUNT):
+        strResult += '      <' + causes[idx] + '>\n'
+        strResult += GetRetiredPagesByCause(handle, idx)
+        strResult += '      </' + causes[idx] + '>\n'
+
+    strResult += '      <pending_retirement>'
+    try:
+        if NVML_FEATURE_DISABLED == nvmlDeviceGetRetiredPagesPendingStatus(handle):
+            strResult += "No"
+        else:
+            strResult += "Yes"
+    except NVMLError as err:
+        strResult += handleError(err)
+    strResult += '</pending_retirement>\n'
+    return strResult
+    
 def StrGOM(mode):
     if mode == NVML_GOM_ALL_ON:
         return "All On";
@@ -130,6 +164,7 @@ def GetClocksThrottleReasons(handle):
     throttleReasons = [
             [nvmlClocksThrottleReasonGpuIdle,           "clocks_throttle_reason_gpu_idle"],
             [nvmlClocksThrottleReasonUserDefinedClocks, "clocks_throttle_reason_user_defined_clocks"],
+            [nvmlClocksThrottleReasonApplicationsClocksSetting, "clocks_throttle_reason_applications_clocks_setting"],
             [nvmlClocksThrottleReasonSwPowerCap,        "clocks_throttle_reason_sw_power_cap"],
             [nvmlClocksThrottleReasonHwSlowdown,        "clocks_throttle_reason_hw_slowdown"],
             [nvmlClocksThrottleReasonUnknown,           "clocks_throttle_reason_unknown"]
@@ -142,11 +177,12 @@ def GetClocksThrottleReasons(handle):
         clocksThrottleReasons = nvmlDeviceGetCurrentClocksThrottleReasons(handle);
         strResult += '    <clocks_throttle_reasons>\n'
         for (mask, name) in throttleReasons:
-            if (mask & supportedClocksThrottleReasons):
-                val = "Active" if mask & clocksThrottleReasons else "Not Active";
-            else:
-                val = handleError(NVML_ERROR_NOT_SUPPORTED);
-            strResult += "      <%s>%s</%s>\n" % (name, val, name);
+            if (name != "clocks_throttle_reason_user_defined_clocks"):
+                if (mask & supportedClocksThrottleReasons):
+                    val = "Active" if mask & clocksThrottleReasons else "Not Active";
+                else:
+                    val = handleError(NVML_ERROR_NOT_SUPPORTED);
+                strResult += "      <%s>%s</%s>\n" % (name, val, name);
         strResult += '    </clocks_throttle_reasons>\n'
     except NVMLError as err:
         strResult += '    <clocks_throttle_reasons>%s</clocks_throttle_reasons>\n' % (handleError(err));
@@ -165,12 +201,12 @@ def handleError(err):
 #######
 def XmlDeviceQuery():
 
+    strResult = ''
     try:
         #
         # Initialize NVML
         #
         nvmlInit()
-        strResult = ''
 
         strResult += '<?xml version="1.0" ?>\n'
         strResult += '<!DOCTYPE nvidia_smi_log SYSTEM "nvsmi_device_v4.dtd">\n'
@@ -191,6 +227,22 @@ def XmlDeviceQuery():
             
             strResult += '    <product_name>' + nvmlDeviceGetName(handle) + '</product_name>\n'
             
+            brandNames = {NVML_BRAND_UNKNOWN :  "GeForce",
+                          NVML_BRAND_QUADRO  :  "Quadro",
+                          NVML_BRAND_TESLA   :  "Tesla",
+                          NVML_BRAND_NVS     :  "NVS",
+                          NVML_BRAND_GRID    :  "Grid",
+            }
+
+            try:
+                # if nvmlDeviceGetBrand() succeeds it is guaranteed to be in the dictionary
+                brandName = brandNames[nvmlDeviceGetBrand(handle)]
+            except NVMLError as err:
+                brandName = handleError(err)
+
+
+            strResult += '    <product_brand>' + brandName + '</product_brand>\n'
+                
             try:
                 state = ('Enabled' if (nvmlDeviceGetDisplayMode(handle) != 0) else 'Disabled')
             except NVMLError as err:
@@ -198,12 +250,33 @@ def XmlDeviceQuery():
             
             strResult += '    <display_mode>' + state + '</display_mode>\n'
             
+            try:
+                state = ('Enabled' if (nvmlDeviceGetDisplayActive(handle) != 0) else 'Disabled')
+            except NVMLError as err:
+                state = handleError(err)
+            
+            strResult += '    <display_active>' + state + '</display_active>\n'
+            
             try:
                 mode = 'Enabled' if (nvmlDeviceGetPersistenceMode(handle) != 0) else 'Disabled'
             except NVMLError as err:
                 mode = handleError(err)
             
             strResult += '    <persistence_mode>' + mode + '</persistence_mode>\n'
+            
+            try:
+                mode = 'Enabled' if (nvmlDeviceGetAccountingMode(handle) != 0) else 'Disabled'
+            except NVMLError as err:
+                mode = handleError(err)
+            
+            strResult += '    <accounting_mode>' + mode + '</accounting_mode>\n'
+           
+            try:
+                bufferSize = str(nvmlDeviceGetAccountingBufferSize(handle))
+            except NVMLError as err:
+                bufferSize = handleError(err)
+            
+            strResult += '    <accounting_mode_buffer_size>' + bufferSize + '</accounting_mode_buffer_size>\n'
                 
             strResult += '    <driver_model>\n'
 
@@ -236,6 +309,13 @@ def XmlDeviceQuery():
 
             strResult += '    <uuid>' + uuid + '</uuid>\n'
             
+            try:
+                minor_number = nvmlDeviceGetMinorNumber(handle)
+            except NVMLError as err:
+                minor_number = handleError(err)
+
+            strResult += '    <minor_number>' + str(minor_number) + '</minor_number>\n'
+            
             try:
                 vbios = nvmlDeviceGetVbiosVersion(handle)
             except NVMLError as err:
@@ -243,6 +323,30 @@ def XmlDeviceQuery():
 
             strResult += '    <vbios_version>' + vbios + '</vbios_version>\n'
 
+            try:
+                multiGpuBool = nvmlDeviceGetMultiGpuBoard(handle)
+            except NVMLError as err:
+                multiGpuBool = handleError(err);
+
+            if multiGpuBool == "N/A":
+                strResult += '    <multigpu_board>' + 'N/A' + '</multigpu_board>\n'
+            elif multiGpuBool:
+                strResult += '    <multigpu_board>' + 'Yes' + '</multigpu_board>\n'
+            else:
+                strResult += '    <multigpu_board>' + 'No' + '</multigpu_board>\n'
+
+            try:
+                boardId = nvmlDeviceGetBoardId(handle)
+            except NVMLError as err:
+                boardId = handleError(err)
+
+            try:
+                hexBID = "0x%x" % boardId
+            except: 
+                hexBID = boardId
+
+            strResult += '    <board_id>' + hexBID + '</board_id>\n'
+
             strResult += '    <inforom_version>\n'
             
             try:
@@ -336,6 +440,30 @@ def XmlDeviceQuery():
 
             strResult += '        </link_widths>\n'
             strResult += '      </pci_gpu_link_info>\n'
+
+
+            strResult += '      <pci_bridge_chip>\n'
+
+            try:
+                bridgeHierarchy = nvmlDeviceGetBridgeChipInfo(handle)
+                bridge_type = ''
+                if bridgeHierarchy.bridgeChipInfo[0].type == 0:
+                    bridge_type += 'PLX'
+                else:
+                    bridge_type += 'BR04'                    
+                strResult += '        <bridge_chip_type>' + bridge_type + '</bridge_chip_type>\n'
+
+                if bridgeHierarchy.bridgeChipInfo[0].fwVersion == 0:
+                    strFwVersion = 'N/A'
+                else:
+                    strFwVersion = '%08X' % (bridgeHierarchy.bridgeChipInfo[0].fwVersion)
+                strResult += '        <bridge_chip_fw>%s</bridge_chip_fw>\n' % (strFwVersion)
+            except NVMLError as err:
+                strResult += '        <bridge_chip_type>' + handleError(err) + '</bridge_chip_type>\n'
+                strResult += '        <bridge_chip_fw>' + handleError(err) + '</bridge_chip_fw>\n'
+
+            # Add additional code for hierarchy of bridges for Bug # 1382323                
+            strResult += '      </pci_bridge_chip>\n'
             strResult += '    </pci>\n'
 
             try:
@@ -355,21 +483,37 @@ def XmlDeviceQuery():
 
             try:
                 memInfo = nvmlDeviceGetMemoryInfo(handle)
-                mem_total = str(memInfo.total / 1024 / 1024) + ' MB'
-                mem_used = str(memInfo.used / 1024 / 1024) + ' MB'
-                mem_free = str(memInfo.total / 1024 / 1024 - memInfo.used / 1024 / 1024) + ' MB'
+                mem_total = str(memInfo.total / 1024 / 1024) + ' MiB'
+                mem_used = str(memInfo.used / 1024 / 1024) + ' MiB'
+                mem_free = str(memInfo.total / 1024 / 1024 - memInfo.used / 1024 / 1024) + ' MiB'
             except NVMLError as err:
                 error = handleError(err)
                 mem_total = error
                 mem_used = error
                 mem_free = error
 
-            strResult += '    <memory_usage>\n'
+            strResult += '    <fb_memory_usage>\n'
             strResult += '      <total>' + mem_total + '</total>\n'
             strResult += '      <used>' + mem_used + '</used>\n'
             strResult += '      <free>' + mem_free + '</free>\n'
-            strResult += '    </memory_usage>\n'
+            strResult += '    </fb_memory_usage>\n'
 
+            try:
+                memInfo = nvmlDeviceGetBAR1MemoryInfo(handle)
+                mem_total = str(memInfo.bar1Total / 1024 / 1024) + ' MiB'
+                mem_used = str(memInfo.bar1Used / 1024 / 1024) + ' MiB'
+                mem_free = str(memInfo.bar1Total / 1024 / 1024 - memInfo.bar1Used / 1024 / 1024) + ' MiB'
+            except NVMLError as err:
+                error = handleError(err)
+                mem_total = error
+                mem_used = error
+                mem_free = error
+
+            strResult += '    <bar1_memory_usage>\n'
+            strResult += '      <total>' + mem_total + '</total>\n'
+            strResult += '      <used>' + mem_used + '</used>\n'
+            strResult += '      <free>' + mem_free + '</free>\n'
+            strResult += '    </bar1_memory_usage>\n'
             
             try:
                 mode = nvmlDeviceGetComputeMode(handle)
@@ -400,6 +544,25 @@ def XmlDeviceQuery():
             strResult += '    <utilization>\n'
             strResult += '      <gpu_util>' + gpu_util + '</gpu_util>\n'
             strResult += '      <memory_util>' + mem_util + '</memory_util>\n'
+
+            try:
+                (util_int, ssize) = nvmlDeviceGetEncoderUtilization(handle)
+                encoder_util = str(util_int) + ' %'
+            except NVMLError as err:
+                error = handleError(err)
+                encoder_util = error
+
+            strResult += '      <encoder_util>' + encoder_util + '</encoder_util>\n'
+
+            try:
+                (util_int, ssize) = nvmlDeviceGetDecoderUtilization(handle)
+                decoder_util = str(util_int) + ' %'
+            except NVMLError as err:
+                error = handleError(err)
+                decoder_util = error
+
+            strResult += '      <decoder_util>' + decoder_util + '</decoder_util>\n'
+
             strResult += '    </utilization>\n'
             
             try:
@@ -419,6 +582,10 @@ def XmlDeviceQuery():
             strResult += '    <ecc_errors>\n'
             strResult += GetEccStr(handle)
             strResult += '    </ecc_errors>\n'
+
+            strResult += '    <retired_pages>\n'
+            strResult += GetRetiredPagesStr(handle)
+            strResult += '    </retired_pages>\n'
             
             try:
                 temp = str(nvmlDeviceGetTemperature(handle, NVML_TEMPERATURE_GPU)) + ' C'
@@ -427,6 +594,20 @@ def XmlDeviceQuery():
 
             strResult += '    <temperature>\n'
             strResult += '      <gpu_temp>' + temp + '</gpu_temp>\n'
+
+            try:
+                temp = str(nvmlDeviceGetTemperatureThreshold(handle, NVML_TEMPERATURE_THRESHOLD_SHUTDOWN)) + ' C'
+            except NVMLError as err:
+                temp = handleError(err)
+
+            strResult += '      <gpu_temp_max_threshold>' + temp + '</gpu_temp_max_threshold>\n'
+
+            try:
+                temp = str(nvmlDeviceGetTemperatureThreshold(handle, NVML_TEMPERATURE_THRESHOLD_SLOWDOWN)) + ' C'
+            except NVMLError as err:
+                temp = handleError(err)
+
+            strResult += '      <gpu_temp_slow_threshold>' + temp + '</gpu_temp_slow_threshold>\n'
             strResult += '    </temperature>\n'
 
             strResult += '    <power_readings>\n'
@@ -459,6 +640,15 @@ def XmlDeviceQuery():
             except NVMLError as err:
                 powLimitStr = handleError(err)
             strResult += '      <default_power_limit>' + powLimitStr + '</default_power_limit>\n'
+
+            try:
+                enforcedPowLimit = (nvmlDeviceGetEnforcedPowerLimit(handle) / 1000.0)
+                enforcedPowLimitStr = '%.2f W' % enforcedPowLimit
+            except NVMLError as err:
+                enforcedPowLimitStr = handleError(err)
+
+            strResult += '      <enforced_power_limit>' + enforcedPowLimitStr + '</enforced_power_limit>\n'
+
             try:
                 powLimit = nvmlDeviceGetPowerManagementLimitConstraints(handle)
                 powLimitStrMin = '%.2f W' % (powLimit[0] / 1000.0)
@@ -474,10 +664,10 @@ def XmlDeviceQuery():
 
             strResult += '    <clocks>\n'
             try:
-                graphics = str(nvmlDeviceGetClockInfo(handle, NVML_CLOCK_GRAPHICS))
+                graphics = str(nvmlDeviceGetClockInfo(handle, NVML_CLOCK_GRAPHICS)) + ' MHz'
             except NVMLError as err:
                 graphics = handleError(err)
-            strResult += '      <graphics_clock>' +graphics + ' MHz</graphics_clock>\n'
+            strResult += '      <graphics_clock>' +graphics + '</graphics_clock>\n'
             try:
                 sm = str(nvmlDeviceGetClockInfo(handle, NVML_CLOCK_SM)) + ' MHz'
             except NVMLError as err:
@@ -502,6 +692,19 @@ def XmlDeviceQuery():
                 mem = handleError(err)
             strResult += '      <mem_clock>' + mem + '</mem_clock>\n'
             strResult += '    </applications_clocks>\n'
+            
+            strResult += '    <default_applications_clocks>\n'
+            try:
+                graphics = str(nvmlDeviceGetDefaultApplicationsClock(handle, NVML_CLOCK_GRAPHICS)) + ' MHz'
+            except NVMLError as err:
+                graphics = handleError(err)
+            strResult += '      <graphics_clock>' +graphics + '</graphics_clock>\n'
+            try:
+                mem = str(nvmlDeviceGetDefaultApplicationsClock(handle, NVML_CLOCK_MEM)) + ' MHz'
+            except NVMLError as err:
+                mem = handleError(err)
+            strResult += '      <mem_clock>' + mem + '</mem_clock>\n'
+            strResult += '    </default_applications_clocks>\n'
 
             strResult += '    <max_clocks>\n'
             try:
@@ -520,6 +723,30 @@ def XmlDeviceQuery():
                 mem = handleError(err)
             strResult += '      <mem_clock>' + mem + '</mem_clock>\n'
             strResult += '    </max_clocks>\n'
+            
+            strResult += '    <clock_policy>\n'
+            try:
+                boostedState, boostedDefaultState = nvmlDeviceGetAutoBoostedClocksEnabled(handle)
+                if boostedState == NVML_FEATURE_DISABLED:
+                    autoBoostStr = "Off"
+                else:
+                    autoBoostStr = "On"
+                
+                if boostedDefaultState == NVML_FEATURE_DISABLED:
+                    autoBoostDefaultStr = "Off"
+                else:
+                    autoBoostDefaultStr = "On"
+                
+            except NVMLError_NotSupported:
+                autoBoostStr = "N/A"
+                autoBoostDefaultStr = "N/A"
+            except NVMLError as err:
+                autoBoostStr = handleError(err)
+                autoBoostDefaultStr = handleError(err)
+                pass
+            strResult += '      <auto_boost>' + autoBoostStr + '</auto_boost>\n'
+            strResult += '      <auto_boost_default>' + autoBoostDefaultStr + '</auto_boost_default>\n'
+            strResult += '    </clock_policy>\n'
 
             try:
                 memClocks = nvmlDeviceGetSupportedMemoryClocks(handle)
@@ -561,13 +788,49 @@ def XmlDeviceQuery():
                     if (p.usedGpuMemory == None):
                         mem = 'N\A'
                     else:
-                        mem = '%d MB' % (p.usedGpuMemory / 1024 / 1024)
+                        mem = '%d MiB' % (p.usedGpuMemory / 1024 / 1024)
                     strResult += '      <used_memory>' + mem + '</used_memory>\n'
                     strResult += '    </process_info>\n'
                 
                 strResult += '    </compute_processes>\n'
             except NVMLError as err:
                 strResult += '    <compute_processes>' + handleError(err) + '</compute_processes>\n'
+            
+            try:
+                pids = nvmlDeviceGetAccountingPids(handle)
+                strResult += '    <accounted_processes>\n'
+             
+                for pid in pids :
+                    try:
+                        stats = nvmlDeviceGetAccountingStats(handle, pid) 
+                        gpuUtilization = "%d %%" % stats.gpuUtilization
+                        memoryUtilization = "%d %%" % stats.memoryUtilization
+                        if (stats.maxMemoryUsage == None):
+                            maxMemoryUsage = 'N\A'
+                        else:
+                            maxMemoryUsage = '%d MiB' % (stats.maxMemoryUsage / 1024 / 1024)
+                        time = "%d ms" % stats.time
+                    except NVMLError as err:
+                        if (err.value == NVML_ERROR_NOT_FOUND):
+                            # probably went away
+                            continue
+                        err = handleError(err)
+                        gpuUtilization = err
+                        memoryUtilization = err
+                        maxMemoryUsage = err
+                        time = err
+                    
+                    strResult += '    <accounted_process_info>\n'
+                    strResult += '      <pid>%d</pid>\n' % pid
+                    strResult += '      <gpu_util>' + gpuUtilization + '</gpu_util>\n'
+                    strResult += '      <memory_util>' + memoryUtilization + '</memory_util>\n'
+                    strResult += '      <max_memory_usage>' + maxMemoryUsage+ '</max_memory_usage>\n'
+                    strResult += '      <time>' + time + '</time>\n'
+                    strResult += '    </accounted_process_info>\n'
+                
+                strResult += '    </accounted_processes>\n'
+            except NVMLError as err:
+                strResult += '    <accounted_processes>' + handleError(err) + '</accounted_processes>\n'
 
             strResult += '  </gpu>\n'
             
@@ -580,3 +843,6 @@ def XmlDeviceQuery():
     
     return strResult
 
+# this is not exectued when module is imported
+if __name__ == "__main__":
+    print(XmlDeviceQuery())
